/* Xpede v0.01
 * Copyright 1996 P. Warden (wardenp@cs.man.ac.uk)
 * This software may be freely copied and used, but no warranty is given
 */

/* include files *********************************************************/

/* include the files needed for Xlib */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

/* include files for io and stdlib for the rand function */
#include <stdio.h>
#include <stdlib.h>

/* definitions and enumerations ******************************************/

/* length of snake's postion history buffer */
#define HISTLENGTH (1024)

/* application's event mask */
#define MASK ExposureMask | KeyPressMask | StructureNotifyMask | FocusChangeMask

/* possible values for the playing area map */
enum mtype { EMPTY,MUSH,SNAKE,KILL,HDUP,HDRIGHT,HDDOWN,HDLEFT,HEART,UNSET };

/* possible game states */
enum gstate { INGAME=1, PAUSED=2, DYING=4, NEWLEVEL=8, WIN=16 };

/* structure for holding colour map information */
struct cstruct {

  int black;
  int white;
  int red;
  int green;
  int blue;

};

/* structure to hold position information */
struct point {
  int x;
  int y;
};

/* function prototypes *************************************************/

void ld_font(XFontStruct **);
void getGC(Window ,GC *,XFontStruct *);
void make_image(Window ),setup_mushroom(XImage *),make_map(int ,int ,int);
void setup_empty(XImage *),setup_snake(XImage *),setup_hdup(XImage *);
void setup_hdright(XImage *),setup_hddown(XImage *),setup_hdleft(XImage *);
void setup_killer(XImage *),setup_heads(void),update_map(void),setup_game(void);
void update_game(void),draw_y_line(XImage * ,int , int, int ,int);
void setup_heart(XImage *),update_image_from_map(Window ,GC);
void update_image_dead(Window ,GC),update_image(Window, GC),unset_dmap(void); 
void setup_map_memory(void),free_memory(void),print_usage(void);
void update_image_won(Window ,GC),init_game(void);


/* global variables ****************************************************/

/* variable to hold the game's current state */
enum gstate gamestate;

/* variable to store closest pixel values to desired colours */
struct cstruct cvals;

/* variables to hold the current display and screen */
Display *display;
int screen;

/* program name (from argv[0]) */
static char *pname;

/* pointers to the various x format images used in this program */
XImage *empty,*mushroom,*snake,*killer,*hdup,*hdright,*hddown,*hdleft,*heart;

/* pointers to the data areas for all the images */
char *emptyd,*mushroomd,*snaked,*killerd,*hdupd,*hdrightd,*hddownd,*hdleftd,
*heartd;

/* pointer to permanent map */
enum mtype *map;

/* pointer to map used for display purposes, & which bank is currently used */
enum mtype *dmap;
int cdmap;

/* direction snake is heading in */
enum mtype sdir;

/* current position of snake's head */
struct point spos;

/* buffer for past positions of snake's head, & current buffer offset */
struct point shist[HISTLENGTH];
int shistoff;

/* length of snake */
int slength;

/* bit depth all graphics are to be created in */
int depth;

/* counter for the snake death animation */
int dcount;

/* size of game area, in 16 pixel units */
int width=32,height=32;

/* difficulty level */
int difficulty=0;

/* number of mushrooms left */
int mushnum;

/* can you wrap around? */
int walls=False;

/* main code **************************************************************/

/* function to control the main execution of the program */

void main(int argc, char *argv[]) {

  Window win; /* id of window */
  int wwidth,wheight,x,y,border_width=4; /* window geometry information */
  /* title bar string */
  char *window_title="XCentipede: keys are z,x & q  use -s to alter speed";
  XSizeHints size_hints; /* desired window size */
  XWMHints wm_hints; /* general hints for the window manager */
  XClassHint class_hints; /* class hints for the window manager */
  XTextProperty window_title_property; /* window name as a property */

  XEvent event; /* structure to hold events */
  GC gc; /* holds the graphics context id */
  XFontStruct *font; /* handle to the current font */
  char *display_name=NULL; /* current display name */
  
  char buffer[20]; /* buffer for decoding key events */
  int bufsize=20; /* buffer size */
  KeySym key; /* key symbol for decoding key events */
  XComposeStatus compose; /* needed for decoding key events */
  char aclkkey='x',clkkey='z',quitkey='q'; /* player`s keys */
  
  int ltime; /* needed to control the game's speed */
  int speed=1000; /* factor to slow game by */
  int count; /* counter for arg parsing loop */

  /* set program name to that given on command line */
  pname=argv[0];

  /* parse the command line arguments */
  for (count=1; count<argc; count++) {

    switch (*(argv[count]+1)) {

      case 'w': sscanf(argv[count]+2,"%d",&width); break;

      case 'h': sscanf(argv[count]+2,"%d",&height); break;

      case 'a': clkkey=tolower(*(argv[count]+2)); break;

      case 'c': aclkkey=tolower(*(argv[count]+2)); break;

      case 'q': quitkey=tolower(*(argv[count]+2)); break;

      case 's': sscanf(argv[count]+2,"%d",&speed); break;

      case 'd': sscanf(argv[count]+2,"%d",&difficulty); break;

      case 't': walls=True; break;

      default:
       (void) fprintf(stderr, "%s: Option '%s' not understood.\n", pname,argv[count]);
       print_usage();
       break;
    }
  }
   
/*  sprintf(window_title,"%s  keys are '%c', '%c' and '%c' to quit\0",
  pname,aclkkey,clkkey,quitkey); */

  /* connect to the X server, or report an error if not possible */
  if ((display=XOpenDisplay(display_name))==NULL) {
    (void) fprintf(stderr,"%s: can't connect to X server %s.\n",
    pname,display_name);
    exit(0); }

  /* get screen number */
  screen=DefaultScreen(display);

  /* get bit depth for graphics */
  depth=DefaultDepth(display,screen);

  /* set window position */
  x=0;
  y=0;

  /* set window size */
  wwidth=width*16;
  wheight=height*16;

  /* create window on server */
  win=XCreateWindow(display,RootWindow(display,screen),x,y,wwidth,wheight,
  border_width,depth,InputOutput,DefaultVisual(display,screen),0,0);

  /* set size hints for the window manager */
  /* note, this asks the window manager to create a window of an exact size */
  /* if this request is not honoured, it may cause problems playing the game */

  size_hints.flags=PPosition | PSize | PMinSize | PMaxSize;
  size_hints.min_width=wwidth;
  size_hints.min_height=wheight;
  size_hints.max_width=wwidth;
  size_hints.max_height=wheight;

  /* create window_title_property */
  if (XStringListToTextProperty(&window_title,1,&window_title_property)==0)
 {
    (void) fprintf(stderr,"%s: structure window_title_property can't be allocated.\n",pname); }

  /* set various hints */
  wm_hints.initial_state=NormalState;
  wm_hints.input=True;
  wm_hints.flags=StateHint | InputHint;
  class_hints.res_name=pname;
  class_hints.res_class="Snake";

  /* pass hints to the window manager */
  XSetWMProperties(display,win,&window_title_property,NULL,argv,argc,
  &size_hints,&wm_hints,&class_hints);

  /* select wanted events */
  XSelectInput(display,win,MASK);

  /* load font */
  ld_font(&font);

  /* create the graphics context */
  getGC(win,&gc,font);

  /* map window onto display */
  XMapWindow(display,win);

  /* create the game's graphics */
  make_image(win);

  /* initialise the games state */
  gamestate=PAUSED | NEWLEVEL | INGAME;

  /* initialise various game things */
  init_game();

  /* main game loop */
  while (True) {

  /* reset timing counter */
    ltime=0;

    /* branch control according to game state */

    /* set up new game if NEWLEVEL is set in gamestate, & clear flag */
    if (gamestate & NEWLEVEL) { setup_game(); gamestate&=~NEWLEVEL; }

    /* only do anything if PAUSED is false in gamestate */
    if (!(gamestate & PAUSED)) {

      /* if the snake has just died, do special animation */
      if (gamestate & DYING) {
        update_image_dead(win,gc);
        dcount++;

        /* if animation has reached the edge of the screen, restart game */
        if (dcount>((width>height)?(width/2):(height/2))) {
          dcount=0; gamestate|=NEWLEVEL; gamestate&=~DYING; }

      } else if (gamestate & WIN) {
        update_image_won(win,gc);
        dcount++;

        /* if animation has reached the edge of the screen, restart game */
        if (dcount>((width>height)?(width/2):(height/2))) {
          dcount=0; gamestate|=NEWLEVEL; gamestate&=~WIN; }

      } else {
        if (gamestate & INGAME) {
          update_game();
          update_image(win,gc);
        }
      }
    }

    /* this loop here to regulate the game's speed */
    do {

      /* while more events remain, grab an event and process it */
      while (XCheckMaskEvent(display,MASK,&event)) {

        /* switch on event type */
        switch (event.type) {

          case Expose:
            if (event.xexpose.count!=0) break; /*more exposes to come*/
            unset_dmap(); /* clear to indicate all must be redrawn */
            update_image(win,gc); /* draw graphics */
            break;

          case ConfigureNotify:
            /* to be filled in! */
            break;

          case KeyPress:
            /* work out actual key pressed as a string */
            XLookupString(&(event.xkey),buffer,bufsize,&key,&compose);

            /* change direction if it's a key we're interested in */
            if (tolower(buffer[0])==aclkkey) sdir=((sdir+1)%4)+4;
            if (tolower(buffer[0])==clkkey) sdir=((sdir-1)%4)+4;

            /* quit if it's the quit key */
            if (tolower(buffer[0])==quitkey) {
              free_memory(); /* free all malloc'd memory */
              XUnloadFont(display,font->fid);
              XFreeGC(display,gc);
              XCloseDisplay(display);
              exit(1);
              break; }

          case FocusIn:
            gamestate&=~PAUSED; break;

          case FocusOut:
            gamestate|=PAUSED; break;  

          default:
            /* do nothing with any other events */
            break;
        } /* end of event switch */
      } /* end of event while */

      /* no more events to process */
      /* increment timing regulator, and repeat event while loop to kill time */
      ltime++;
    } while (ltime<speed); /* don't end until looped speed times */

  } /* end main game loop */

} /* end of main function */

/* function to load the font for x */
void ld_font(XFontStruct **font) {

  char *fontname="9x15"; /* name of font to load */

  if ((*font=XLoadQueryFont(display,fontname))==NULL) {
    (void) fprintf(stderr,"%s: cannot open font '9x15'.\n",pname);
    exit(-1);
  }
}

/* function to make and return a graphics context */
void getGC(Window win, GC *gc, XFontStruct *font) {

  unsigned long valuemask = 0; /* ignore XGCvalues and use defaults */
  XGCValues values;
  unsigned int line_width = 1;
  int line_style = LineSolid;
  int cap_style = CapRound;
  int join_style = JoinRound;
  int dash_offset = 0;
  static char dash_list[] = {12, 24};
  int list_length = 2;

  /* Create default Graphics Context */
  *gc = XCreateGC(display, win, valuemask, &values);

  /* specify font */
  XSetFont(display, *gc, font->fid);

  /* specify black foreground */
  XSetForeground(display, *gc, BlackPixel(display,screen));

  /* set line attributes */
  XSetLineAttributes(display, *gc, line_width, line_style, 
			cap_style, join_style);

  /* set dashes */
  XSetDashes(display, *gc, dash_offset, dash_list, list_length);

}

/* routine to set up all the graphics related things */
void make_image(Window win) {

  XWindowAttributes a; /* needed to get the colormap */
  XColor b,c; /* for the return values of allocated colours */

  /* get the window attributes, so you can get the colour map */
  XGetWindowAttributes(display,win,&a);

  /* allocate memory for the image data */
  emptyd=(char *) malloc(sizeof(char)*4*16*16);
  mushroomd=(char *) malloc(sizeof(char)*4*16*16);
  snaked=(char *) malloc(sizeof(char)*4*16*16);
  killerd=(char *) malloc(sizeof(char)*4*16*16);
  hdupd=(char *) malloc(sizeof(char)*4*16*16);
  hdrightd=(char *) malloc(sizeof(char)*4*16*16);
  hddownd=(char *) malloc(sizeof(char)*4*16*16);
  hdleftd=(char *) malloc(sizeof(char)*4*16*16);
  heartd=(char *) malloc(sizeof(char)*4*16*16);

  /* make sure the memory has been properly allocated */
  if (emptyd==NULL || mushroomd==NULL || snaked==NULL || killerd==NULL ||
      hdupd==NULL || hdrightd==NULL || hddownd==NULL || hdleftd==NULL ||
      heartd==NULL) {
    (void) fprintf(stderr,"%s: Unable to allocate memory.\n",pname);
    exit(1); }

  /* create the blank images */

  empty=XCreateImage(display,a.visual,depth,ZPixmap,0,emptyd,16,16,8,0);
  mushroom=XCreateImage(display,a.visual,depth,ZPixmap,0,mushroomd,16,16,8,0);
  snake=XCreateImage(display,a.visual,depth,ZPixmap,0,snaked,16,16,8,0);
  killer=XCreateImage(display,a.visual,depth,ZPixmap,0,killerd,16,16,8,0);
  hdup=XCreateImage(display,a.visual,depth,ZPixmap,0,hdupd,16,16,8,0);
  hdright=XCreateImage(display,a.visual,depth,ZPixmap,0,hdrightd,16,16,8,0);
  hddown=XCreateImage(display,a.visual,depth,ZPixmap,0,hddownd,16,16,8,0);
  hdleft=XCreateImage(display,a.visual,depth,ZPixmap,0,hdleftd,16,16,8,0);
  heart=XCreateImage(display,a.visual,depth,ZPixmap,0,heartd,16,16,8,0);

  /* get the closest colours available to the ones wanted */

  XAllocNamedColor(display,a.colormap,"black",&b,&c);
  cvals.black=b.pixel;
  XAllocNamedColor(display,a.colormap,"white",&b,&c);
  cvals.white=b.pixel;
  XAllocNamedColor(display,a.colormap,"red",&b,&c);
  cvals.red=b.pixel;
  XAllocNamedColor(display,a.colormap,"green",&b,&c);
  cvals.green=b.pixel;
  XAllocNamedColor(display,a.colormap,"blue",&b,&c);
  cvals.blue=b.pixel;

  /* ensure graphics can be seen on a monochrome screen */
  if (depth==1) {

    cvals.black=BlackPixel(display,screen);
    cvals.white=WhitePixel(display,screen);
    cvals.red=WhitePixel(display,screen);
    cvals.green=WhitePixel(display,screen);
    cvals.blue=WhitePixel(display,screen);

  }

  /* draw the graphics onto the images */

  setup_empty(empty);
  setup_mushroom(mushroom);
  setup_snake(snake);
  setup_killer(killer);
  setup_heads();
  setup_heart(heart);

}

/* function to draw mushroom graphic */
void setup_mushroom(XImage *animage) {

  int x,y,pixval;

  /* set image background to black */
  pixval=cvals.black;
  for (y=0; y<16; y++) {
    for (x=0; x<16; x++) {
      XPutPixel(animage,x,y,pixval);
    }
  }

  /* draw stem in closest we have to green */
  pixval=cvals.green;
  for (y=9; y<15; y++) {
    for (x=4; x<10; x++) {
      XPutPixel(animage,x,y,pixval);
    }
  }

  /* draw the 'dome' of the mushroom */
  y=8;
  for (x=1;x<13;x++) XPutPixel(animage,x,y,pixval);
  y=7;
  for (x=1;x<13;x++) XPutPixel(animage,x,y,pixval);
  y=6;
  for (x=2;x<12;x++) XPutPixel(animage,x,y,pixval);
  y=5;
  for (x=2;x<12;x++) XPutPixel(animage,x,y,pixval);
  y=4;
  for (x=3;x<11;x++) XPutPixel(animage,x,y,pixval);
  y=3;
  for (x=4;x<10;x++) XPutPixel(animage,x,y,pixval);
  y=2;
  for (x=6;x<8;x++) XPutPixel(animage,x,y,pixval);


}

/* function to set up the map, with numbers of each object to be placed
passed to it */

void make_map(int mush,int death,int hearts) {

  int x,y,p,free;
  enum mtype mval; 

  /* setup random seed */
  p=rand();

  /* set initial number of free grid spaces */
  free=width*height; 

  /* set the whole map to be EMPTY */
  for (y=0; y<height; y++) {
    for (x=0; x<width; x++) {
    mval = EMPTY;
    *(map+(x*height)+y)=mval;
   }
  }

  /* unset the display map */
  unset_dmap();

  /* populate the world with mushrooms */
  while (mush>0 && free>5) {

    /* set the x and y positions to random values within the map */
    x=(p<0) ? (-p)%width : p%width;
    p^=(p<<13) | (p>>19);
    p+=1;
    y=(p<0) ? (-p)%height : p%height;

    /* if map point is empty, put a mushroom there, decrement the number
still needing to be placed by one, and decrement the number of free spaces */
    if (*(map+(x*height)+y)==EMPTY) {*(map+(x*height)+y)=MUSH;mush--;free--;}

    p^=(p<<13) | (p>>19);
    p+=1;

  }

  /* sprinkle the world with skull tokens */
  while (death>0 && free>5) {

    /* all as above */
    x=(p<0) ? (-p)%width : p%width;
    p^=(p<<13) | (p>>19);
    p+=1;
    y=(p<0) ? (-p)%height : (p)%height;

    if (*(map+(x*height)+y)==EMPTY) {*(map+(x*height)+y)=KILL;death--;free--;}

    p^=(p<<13) | (p>>19);
    p+=1;

  }

  /* sprinkle the world with heart tokens */
  while (hearts>0 && free>5) {

    /* all as above */
    x=(p<0) ? (-p)%width : p%width;
    p^=(p<<13) | (p>>19);
    p+=1;
    y=(p<0) ? (-p)%height : (p/65536)%height;

    if (*(map+(x*height)+y)==EMPTY) {*(map+(x*height)+y)=HEART;hearts--;free--;}

    p^=(p<<13) | (p>>19);
    p+=1;

  }

}
      
/* paints an image completely black */
void setup_empty(XImage *animage) {
  int x,y,pixval;

  pixval=cvals.black;
  for (y=0; y<16; y++) {
    for (x=0; x<16; x++) {
      XPutPixel(animage,x,y,pixval);
    }
  }

}
      
/* draws snake body onto an image */  
void setup_snake(XImage *animage) {
  int x,y,pixval;

  /* blank entire image to blue */
  pixval=cvals.blue;
  for (y=0; y<16; y++) {
    for (x=0; x<16; x++) {
      XPutPixel(animage,x,y,pixval);
    }
  }

  /* trim off corners, repeating in each corner of the image */
  pixval=cvals.black;

  x=0;
  for (y=0; y<6; y++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
    XPutPixel(animage,x,15-y,pixval);
    XPutPixel(animage,15-x,15-y,pixval);
  }

  x=1;
  for (y=0; y<4; y++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
    XPutPixel(animage,x,15-y,pixval);
    XPutPixel(animage,15-x,15-y,pixval);
  }

  x=2;
  for (y=0; y<3; y++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
    XPutPixel(animage,x,15-y,pixval);
    XPutPixel(animage,15-x,15-y,pixval);
  }

  x=3;
  for (y=0; y<2; y++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
    XPutPixel(animage,x,15-y,pixval);
    XPutPixel(animage,15-x,15-y,pixval);
  }

  x=4;
  for (y=0; y<1; y++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
    XPutPixel(animage,x,15-y,pixval);
    XPutPixel(animage,15-x,15-y,pixval);
  }

  x=5;
  for (y=0; y<1; y++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
    XPutPixel(animage,x,15-y,pixval);
    XPutPixel(animage,15-x,15-y,pixval);
  }

}

/* draw the snakes eyes and nose, pointing upwards */
void setup_hdup(XImage *animage) {

  int x,y,pixval;

  /* draw the whites of the eyes */  
  pixval=cvals.white;

  XPutPixel(animage,4,4,pixval);
  XPutPixel(animage,5,4,pixval);
  XPutPixel(animage,3,5,pixval);
  XPutPixel(animage,6,5,pixval);
  XPutPixel(animage,3,6,pixval);
  XPutPixel(animage,6,6,pixval);
  XPutPixel(animage,4,7,pixval);
  XPutPixel(animage,5,7,pixval);

  XPutPixel(animage,11,4,pixval);
  XPutPixel(animage,10,4,pixval);
  XPutPixel(animage,12,5,pixval);
  XPutPixel(animage,9,5,pixval);
  XPutPixel(animage,12,6,pixval);
  XPutPixel(animage,9,6,pixval);
  XPutPixel(animage,11,7,pixval);
  XPutPixel(animage,10,7,pixval);

  /* draw the eye's pupils */
  pixval=cvals.black;

  XPutPixel(animage,4,5,pixval);
  XPutPixel(animage,5,5,pixval);
  XPutPixel(animage,4,6,pixval);
  XPutPixel(animage,5,6,pixval);

  XPutPixel(animage,11,5,pixval);
  XPutPixel(animage,10,5,pixval);
  XPutPixel(animage,11,6,pixval);
  XPutPixel(animage,10,6,pixval);

  /* draw the nose */
  pixval=cvals.red;

  y=0;
  for (x=6; x<10; x++) {
    XPutPixel(animage,x,y,pixval);
  }

  y=1;
  for (x=6; x<10; x++) {
    XPutPixel(animage,x,y,pixval);
  }

  y=2;
  for (x=7; x<9; x++) {
    XPutPixel(animage,x,y,pixval);
  }
}

/* draw the eyes and the nose of the snake, pointing right */
void setup_hdright(XImage *animage) {

  int x,y,pixval;

  /* draw the whites of the eyes */  
  pixval=cvals.white;

  XPutPixel(animage,11,4,pixval);
  XPutPixel(animage,11,5,pixval);
  XPutPixel(animage,10,3,pixval);
  XPutPixel(animage,10,6,pixval);
  XPutPixel(animage,9,3,pixval);
  XPutPixel(animage,9,6,pixval);
  XPutPixel(animage,8,4,pixval);
  XPutPixel(animage,8,5,pixval);

  XPutPixel(animage,11,11,pixval);
  XPutPixel(animage,11,10,pixval);
  XPutPixel(animage,10,12,pixval);
  XPutPixel(animage,10,9,pixval);
  XPutPixel(animage,9,12,pixval);
  XPutPixel(animage,9,9,pixval);
  XPutPixel(animage,8,11,pixval);
  XPutPixel(animage,8,10,pixval);

  /* draw the pupils */
  pixval=cvals.black;

  XPutPixel(animage,10,4,pixval);
  XPutPixel(animage,10,5,pixval);
  XPutPixel(animage,9,4,pixval);
  XPutPixel(animage,9,5,pixval);

  XPutPixel(animage,10,11,pixval);
  XPutPixel(animage,10,10,pixval);
  XPutPixel(animage,9,11,pixval);
  XPutPixel(animage,9,10,pixval);

  /* draw the nose */
  pixval=cvals.red;

  x=15;
  for (y=6; y<10; y++) {
    XPutPixel(animage,x,y,pixval);
  }

  x=14;
  for (y=6; y<10; y++) {
    XPutPixel(animage,x,y,pixval);
  }

  x=13;
  for (y=7; y<9; y++) {
    XPutPixel(animage,x,y,pixval);
  }
}

/* draws the snakes eyes and nose, pointing down */
void setup_hddown(XImage *animage) {

  int x,y,pixval;

  /* draws the whites of the eyes */  
  pixval=cvals.white;

  XPutPixel(animage,4,11,pixval);
  XPutPixel(animage,5,11,pixval);
  XPutPixel(animage,3,10,pixval);
  XPutPixel(animage,6,10,pixval);
  XPutPixel(animage,3,9,pixval);
  XPutPixel(animage,6,9,pixval);
  XPutPixel(animage,4,8,pixval);
  XPutPixel(animage,5,8,pixval);

  XPutPixel(animage,11,11,pixval);
  XPutPixel(animage,10,11,pixval);
  XPutPixel(animage,12,10,pixval);
  XPutPixel(animage,9,10,pixval);
  XPutPixel(animage,12,9,pixval);
  XPutPixel(animage,9,9,pixval);
  XPutPixel(animage,11,8,pixval);
  XPutPixel(animage,10,8,pixval);

  /* draws the pupils */
  pixval=cvals.black;

  XPutPixel(animage,4,10,pixval);
  XPutPixel(animage,5,10,pixval);
  XPutPixel(animage,4,9,pixval);
  XPutPixel(animage,5,9,pixval);

  XPutPixel(animage,11,10,pixval);
  XPutPixel(animage,10,10,pixval);
  XPutPixel(animage,11,9,pixval);
  XPutPixel(animage,10,9,pixval);

  /* draws the nose */
  pixval=cvals.red;

  y=15;
  for (x=6; x<10; x++) {
    XPutPixel(animage,x,y,pixval);
  }

  y=14;
  for (x=6; x<10; x++) {
    XPutPixel(animage,x,y,pixval);
  }

  y=13;
  for (x=7; x<9; x++) {
    XPutPixel(animage,x,y,pixval);
  }
}

/* draws the snake's eyes and nose, pointing left */
void setup_hdleft(XImage *animage) {

  int x,y,pixval;

  /* draws the whites of the eyes */  
  pixval=cvals.white;

  XPutPixel(animage,4,4,pixval);
  XPutPixel(animage,4,5,pixval);
  XPutPixel(animage,5,3,pixval);
  XPutPixel(animage,5,6,pixval);
  XPutPixel(animage,6,3,pixval);
  XPutPixel(animage,6,6,pixval);
  XPutPixel(animage,7,4,pixval);
  XPutPixel(animage,7,5,pixval);

  XPutPixel(animage,4,11,pixval);
  XPutPixel(animage,4,10,pixval);
  XPutPixel(animage,5,12,pixval);
  XPutPixel(animage,5,9,pixval);
  XPutPixel(animage,6,12,pixval);
  XPutPixel(animage,6,9,pixval);
  XPutPixel(animage,7,11,pixval);
  XPutPixel(animage,7,10,pixval);

  /* draws the pupils */
  pixval=cvals.black;

  XPutPixel(animage,5,4,pixval);
  XPutPixel(animage,5,5,pixval);
  XPutPixel(animage,6,4,pixval);
  XPutPixel(animage,6,5,pixval);

  XPutPixel(animage,5,11,pixval);
  XPutPixel(animage,5,10,pixval);
  XPutPixel(animage,6,11,pixval);
  XPutPixel(animage,6,10,pixval);

  /* draws the nose */
  pixval=cvals.red;

  x=0;
  for (y=6; y<10; y++) {
    XPutPixel(animage,x,y,pixval);
  }

  x=1;
  for (y=6; y<10; y++) {
    XPutPixel(animage,x,y,pixval);
  }

  x=2;
  for (y=7; y<9; y++) {
    XPutPixel(animage,x,y,pixval);
  }
}

/* draws the skull onto an image */
void setup_killer(XImage *animage) {
  int x,y,pixval;

  /* blanks the image to black */
  pixval=cvals.black;
  for (y=0; y<16; y++) {
    for (x=0; x<16; x++) {
      XPutPixel(animage,x,y,pixval);
    }
  }

  /* draws the skull */
  pixval=cvals.white;

  y=2;
  for (x=5; x<8; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  y=3;
  for (x=4; x<8; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  y=4;
  for (x=3; x<5; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  x=8;
  XPutPixel(animage,x,y,pixval);
  XPutPixel(animage,15-x,y,pixval);

  y=5;
  for (x=3; x<5; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  x=8;
  XPutPixel(animage,x,y,pixval);
  XPutPixel(animage,15-x,y,pixval);

  y=6;
  for (x=4; x<8; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  y=7;
  for (x=5; x<7; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  y=8;
  for (x=6; x<8; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  y=9;
  for (x=6; x<7; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  y=10;
  for (x=6; x<7; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }

  y=11;
  for (x=6; x<8; x++) {
    XPutPixel(animage,x,y,pixval);
    XPutPixel(animage,15-x,y,pixval);
  }


}

/* organises the drawing of the snake's head rotated to the four positions
onto the right images */
void setup_heads() {

  setup_snake(hdup);
  setup_hdup(hdup);
  setup_snake(hdright);
  setup_hdright(hdright);
  setup_snake(hddown);
  setup_hddown(hddown);
  setup_snake(hdleft);
  setup_hdleft(hdleft);

}

/* combines the snake and the underlying map onto a 'display' map */
void update_map() {

  int x,y,c;

  /* swap the display map bank we're using */
  cdmap=(cdmap+1)%2;

  /* copy the underlying map onto the display map */
  for (y=0; y<height; y++) {
    for (x=0; x<width; x++) {
      *(dmap+(cdmap*height*width)+(x*height)+y)=*(map+(x*height)+y);
    }
  }

  /* overlay the snake onto the display map */
  for (c=1; c<slength; c++) {
    *(dmap+(cdmap*height*width) +
(shist[(shistoff-c+HISTLENGTH)%HISTLENGTH].x*height) +
shist[(shistoff-c+HISTLENGTH)%HISTLENGTH].y ) = SNAKE;
  }

  /* copy the snake's head onto the display map */
  *(dmap+(cdmap*height*width)+(spos.x*height)+spos.y)=sdir;

}

/* setup various game things */
void setup_game() {

  int c;

  /* set snake starting position */
  spos.x=width/2; spos.y=height/2;

  /* set snake starting direction */
  sdir=HDUP;

  /* setup snake's movement history buffer */
  shistoff=0; slength=5;
  for (c=0; c<=slength; c++) {
    shist[(shistoff-c+HISTLENGTH)%HISTLENGTH].x=spos.x;
    shist[(shistoff-c+HISTLENGTH)%HISTLENGTH].y=spos.y+c;
  }
  /* setup the map with mushrooms, skulls and hearts */

  c=((difficulty*.3)+1)*20;
  c=(c*width*height)/256;
  mushnum=c;

  make_map(c,c/4,c/40);

}

/* update various game things */
void update_game() {

  enum mtype newval;
  int count;

  /* change snake's postition, dependant on snakes direction */
  switch (sdir) {

    case HDUP: spos.y--; break;

    case HDRIGHT: spos.x++; break;

    case HDDOWN: spos.y++; break;

    case HDLEFT: spos.x--; break;

  }

  /* move the history buffer forward one place */
  shistoff++;
  shistoff%=HISTLENGTH;

  if (((spos.x<0)||(spos.y<0)||(spos.x>=width)||(spos.y>=height))&&walls) {
    gamestate|=DYING; dcount=0; newval=EMPTY;
    spos.x=(spos.x>=width) ? width-1: spos.x;
    spos.x=(spos.x<0) ? 0: spos.x;
    spos.y=(spos.y>=height) ? height-1: spos.y;
    spos.y=(spos.y<0) ? 0: spos.y;
  }

  /* make sure position is positive */
  spos.x = (spos.x<0) ? spos.x+width : spos.x;
  spos.y = (spos.y<0) ? spos.y+height : spos.y;

  /* wrap around position */
  spos.x%=width;
  spos.y%=height;

  /* store position in history buffer */
  shist[shistoff]=spos;

  /* check to see what snake has landed on */
  switch (*(map+(spos.x*height)+spos.y)) {

    case EMPTY: newval=EMPTY; break; /* do nothing */

   /* increase length, remove mushroom & update number of mushrooms left */
    case MUSH: slength++; newval=EMPTY; mushnum--; break;

    case KILL: gamestate|=DYING; dcount=0; newval=EMPTY; break; /* kill snake */

    case HEART: slength/=2; newval=EMPTY; break; /* reduce snake length */

    default: newval=*(map+(spos.x*height)+spos.y); break; /* do nothing */

  }

  *(map+(spos.x*height)+spos.y)=newval; /* update map */

  /* check history buffer to see if the snake has collided with its own tail */
  for (count=1; count<slength; count++) {

    if (  ((spos.x)==(shist[(shistoff+(HISTLENGTH-count))%HISTLENGTH].x))
       && ((spos.y)==(shist[(shistoff+(HISTLENGTH-count))%HISTLENGTH].y)) ) 
         { gamestate|=DYING; dcount=0; }
  }

  if (mushnum<=0) { gamestate|=WIN; difficulty++; }
    
}

/* draw a line on an image straight down */
void draw_y_line(XImage *animage,int x,int ystart,int yend,int col)
{
  int c;

  for (c=ystart; c<=yend; c++) XPutPixel(animage,x,c,col);

}

/* draw the heart onto an image */
void setup_heart(XImage *animage) {

  int pixval;

  /* clear image to black */
  setup_empty(animage);

  /* draw heart */
  pixval=cvals.red;
  draw_y_line(animage,2,5,7,pixval); draw_y_line(animage,12,5,7,pixval);
  draw_y_line(animage,3,4,9,pixval); draw_y_line(animage,11,4,9,pixval);
  draw_y_line(animage,4,4,11,pixval); draw_y_line(animage,10,4,11,pixval);
  draw_y_line(animage,5,5,12,pixval); draw_y_line(animage,9,5,12,pixval);
  draw_y_line(animage,6,6,14,pixval); draw_y_line(animage,8,6,14,pixval);
  draw_y_line(animage,7,7,15,pixval);

  /* draw highlights onto heart */
  pixval=cvals.white;
  XPutPixel(animage,3,6,pixval); XPutPixel(animage,4,5,pixval);
  XPutPixel(animage,9,6,pixval); XPutPixel(animage,10,5,pixval);

}



/* draw the graphics from the map onto the screen */
void update_image_from_map(Window win,GC gc) {

  int x,y;

  for (y=0;y<(height*16);y+=16) {
    for (x=0;x<(width*16);x+=16) {

      /* only draw if the display map has changed since last time updated */
      if (*(dmap+(cdmap*height*width)  + ((x/16)*height) + (y/16)) != *(dmap+(((cdmap+1)%2)*height*width)+((x/16)*height)+(y/16)) ) {

        switch (*(dmap+(cdmap*height*width)+((x/16)*height)+(y/16))) {

          case EMPTY : XPutImage(display,win,gc,empty,0,0,x,y,16,16); break;
          case MUSH : XPutImage(display,win,gc,mushroom,0,0,x,y,16,16); break;
          case SNAKE : XPutImage(display,win,gc,snake,0,0,x,y,16,16); break;
          case KILL : XPutImage(display,win,gc,killer,0,0,x,y,16,16); break;
          case HDUP : XPutImage(display,win,gc,hdup,0,0,x,y,16,16); break;
          case HDRIGHT : XPutImage(display,win,gc,hdright,0,0,x,y,16,16); break;
          case HDDOWN : XPutImage(display,win,gc,hddown,0,0,x,y,16,16); break;
          case HDLEFT : XPutImage(display,win,gc,hdleft,0,0,x,y,16,16); break;
          case HEART : XPutImage(display,win,gc,heart,0,0,x,y,16,16); break;

        } 
      }
    }
  }

  /* wait until the graphics have appeared */
  XSync(display,False);

}

/* draw graphics after snake has died */
void update_image_dead(Window win, GC gc) {

  int x,y,a,b;

  /* put the normal stuff on the display map */
  update_map();

  /* set size of rectangle of skulls to draw */
  a= (dcount<(width/2)) ? dcount : (width/2);
  b= (dcount<(height/2)) ? dcount : (height/2);

  /* draw the rectangle of skulls */
  for (x=(width/2)-a; x<(width/2)+a; x++) {
    for (y=(height/2)-b; y<(height/2)+b; y++) {
      *(dmap+(cdmap*height*width)+(x*height)+y)=KILL;
    }
  }

  /* draw the display map onto the screen */
  update_image_from_map(win,gc);

}

/* update the display map, and draw it */
void update_image(Window win,GC gc) {

  update_map();
  update_image_from_map(win,gc);

}

/* clear all the display map to UNSET */
void unset_dmap() {

  int x,y;

  for (x=0; x<width; x++) {
    for (y=0; y<height; y++) {
      *(dmap+(cdmap*height*width)+(x*height)+y)=UNSET;
    }
  }

}

/* get the memory needed for the maps */
void setup_map_memory(void) {
 
  map=(enum mtype *) malloc(sizeof(enum mtype)*width*height);
  dmap=(enum mtype *) malloc(sizeof(enum mtype)*2*width*height);

  if ((map==NULL) || (dmap==NULL)) {
    (void) fprintf(stderr,"%s: Unable to allocate memory.\n",pname);
    exit(1); }

}

/* free all allocated memory */
void free_memory(void) {

  /* free image data */
  free(emptyd);
  free(snaked);
  free(mushroomd);
  free(killerd);
  free(hdupd);
  free(hdrightd);
  free(hddownd);
  free(hdleftd);
  free(heartd);

  /* free map data */
  free(map);
  free(dmap);

}

void print_usage(void) {

  printf("Usage: %s [-w<width>] [-h<height>] [-a<anticlockwise key>]",pname);
  printf(" [-c<clockwisekey>] [-q<quitkey>] [-s<slowdown factor>] [-d<level>]\n");

}
  
/* draw graphics after player has won */
void update_image_won(Window win, GC gc) {

  int x,y,a,b;

  /* put the normal stuff on the display map */
  update_map();

  /* set size of rectangle of skulls to draw */
  a= (dcount<(width/2)) ? dcount : (width/2);
  b= (dcount<(height/2)) ? dcount : (height/2);

  /* draw the rectangle of skulls */
  for (x=(width/2)-a; x<(width/2)+a; x++) {
    for (y=(height/2)-b; y<(height/2)+b; y++) {
      *(dmap+(cdmap*height*width)+(x*height)+y)=HEART;
    }
  }

  /* draw the display map onto the screen */
  update_image_from_map(win,gc);

}

/* initialise various game things */

void init_game(void) {

  /* get the memory for the maps */
  setup_map_memory();

}



